// iMSHighSeqCountExample.cpp : This file contains the 'main' function. Program execution begins and ends there.
//

// These are the API header files we will need 
#include "ConnectionList.h"
#include "IMSSystem.h"

#include <iostream>
#include <conio.h>

#include "Supervisors.h"
#include "ImageGen.h"
#include "LibVersion.h"

#define TEST_NUMBER_OF_SEQUENCES (8)      // range 1 - 10
#define TEST_NUMBER_OF_IMAGES    (512)    // range 5 - 32K

// All API classes are defined in the iMS namespace.  For convenience, we can declare this here
using namespace iMS;

// Operating Parameters
double Ch1Amp = 50; // Optimum Power, % 
double Ch2Amp = 50; // Optimum Power, %
double Ch3Amp = 50; // Optimum Power, %
double Ch4Amp = 50; // Optimum Power, %
double DDSlvl = 85; // DDS output level, % (DDSPowerLevel)
double ImgClk = 1000; // Image clock rate, KHz,  

/// Class definition for global access - use this For Single instance communications objects with the iMS
class MyIMSAccess {
public:
	// Global access to single instance SignalPath class
	SignalPath& sigpath() { return this->_sigpath; }
	// Global access to single instance SequenceManager class
	SequenceManager& seqMgr() { return this->_seqMgr; }
	MyIMSAccess(std::shared_ptr<IMSSystem> myiMS) : _sigpath(*myiMS), _seqMgr(*myiMS), thisIMS(myiMS) {}
	~MyIMSAccess() {}
private:
	MyIMSAccess(const MyIMSAccess&);  // Non-copyable
	const MyIMSAccess& operator =(const MyIMSAccess&);  // Non-assignable
	SignalPath _sigpath;
	SequenceManager _seqMgr;
	// Keep a reference to the IMS System to avoid it being deleted before us
	std::shared_ptr<IMSSystem> thisIMS;
};

std::unique_ptr < MyIMSAccess > glbl;

static bool wait_for_image_download(IMSSystem& ims, const Image& img)
{
	ImageDownload* dl = new ImageDownload(ims, img);
	ImageDownloadSupervisor ds;

	dl->ImageDownloadEventSubscribe(ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
	dl->ImageDownloadEventSubscribe(ImageDownloadEvents::DOWNLOAD_ERROR, &ds);

	if (!dl->StartDownload()) {
		std::cout << "Error starting download" << std::endl;
		dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
		dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_ERROR, &ds);
		delete dl;
		return false;
	}

	int timeout = 1000;
	while (ds.Busy()) {
		std::this_thread::sleep_for(std::chrono::milliseconds(5));
		if (timeout-- < 0) {
			std::cout << "Timeout on Image Download." << std::endl;
			return false;
		}
	}

	dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
	dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_ERROR, &ds);

	if (ds.Error()) return false;

	delete dl;
	return true;
}

bool wait_for_sequence_download(IMSSystem& ims, const ImageSequence& seq, bool fast_download = true)
{
	std::unique_ptr<SequenceDownload> dl = std::make_unique<SequenceDownload>(ims, seq);
	SequenceDownloadSupervisor ds;

	dl->SequenceDownloadEventSubscribe(SequenceDownloadEvents::DOWNLOAD_FINISHED, &ds);
	dl->SequenceDownloadEventSubscribe(SequenceDownloadEvents::DOWNLOAD_FAIL_TRANSFER_ABORT, &ds);
	dl->SequenceDownloadEventSubscribe(SequenceDownloadEvents::DOWNLOAD_ERROR, &ds);

	// Confirm that the linked iMS SDK supports the fast download process.
	// If it does, the library itself will check that the iMS firmware also supports it.
	// If either check fails, only the slower download process can be used.
	if (!LibVersion::HasFeature("FAST_SEQUENCE_DOWNLOAD")) {
		fast_download = false;
	}

	if (!dl->Download(fast_download)) {
		std::cout << "Error starting download" << std::endl;
		dl->SequenceDownloadEventUnsubscribe(SequenceDownloadEvents::DOWNLOAD_FINISHED, &ds);
		dl->SequenceDownloadEventUnsubscribe(SequenceDownloadEvents::DOWNLOAD_FAIL_TRANSFER_ABORT, &ds);
		dl->SequenceDownloadEventUnsubscribe(SequenceDownloadEvents::DOWNLOAD_ERROR, &ds);
		return false;
	}

	while (ds.Busy() && fast_download) {
		std::this_thread::sleep_for(std::chrono::milliseconds(50));
	}

	dl->SequenceDownloadEventUnsubscribe(SequenceDownloadEvents::DOWNLOAD_FINISHED, &ds);
	dl->SequenceDownloadEventUnsubscribe(SequenceDownloadEvents::DOWNLOAD_FAIL_TRANSFER_ABORT, &ds);
	dl->SequenceDownloadEventUnsubscribe(SequenceDownloadEvents::DOWNLOAD_ERROR, &ds);

	return !(ds.Error());
}

static void downloadImagesToIMS(std::shared_ptr<iMS::IMSSystem> myiMS, const ImageGroup& grp)
{
	int i = 0;
	std::wcout << "Downloading images" << std::endl;
	for (const auto& img : grp) {
		if ((i++ % 100) == 0) std::cout << ".";
		//std::cout << "Downloading image " << img.Name() << std::endl;
		if (!wait_for_image_download(*myiMS, img)) {
			std::cout << "Aborting.. <press any key>" << std::endl;
			char c;
			std::cin >> c;
			std::exit(-1);
		}
	}
	std::cout << std::endl;
}

static void setInitialConditions()
{
	glbl->sigpath().EnableImagePathCompensation(SignalPath::Compensation::BYPASS, SignalPath::Compensation::BYPASS);
	glbl->sigpath().UpdateDDSPowerLevel(Percent(DDSlvl)); // DDS Level

// Set RF Amplitude to use internal level control (not external modulation)
	glbl->sigpath().SwitchRFAmplitudeControlSource(SignalPath::AmplitudeControl::INDEPENDENT, RFChannel::all);
	// Update per-channel RF Amplitude levels
	glbl->sigpath().UpdateRFAmplitude(SignalPath::AmplitudeControl::INDEPENDENT, Percent(Ch1Amp), RFChannel(1));
	glbl->sigpath().UpdateRFAmplitude(SignalPath::AmplitudeControl::INDEPENDENT, Percent(Ch2Amp), RFChannel(2));
	glbl->sigpath().UpdateRFAmplitude(SignalPath::AmplitudeControl::INDEPENDENT, Percent(Ch3Amp), RFChannel(3));
	glbl->sigpath().UpdateRFAmplitude(SignalPath::AmplitudeControl::INDEPENDENT, Percent(Ch4Amp), RFChannel(4));

	// Use analog frequency outputs to visualise Lissajous figures
	glbl->sigpath().AssignSynchronousOutput(SignalPath::SYNC_SINK::ANLG_A, SignalPath::SYNC_SRC::IMAGE_ANLG_A);
	glbl->sigpath().AssignSynchronousOutput(SignalPath::SYNC_SINK::ANLG_B, SignalPath::SYNC_SRC::IMAGE_ANLG_B);
}

int main(int argc, char* argv[])
{
	SequenceSupervisor seqs;

	//
	// CONNECT TO iMS System
	//
	std::unique_ptr<ConnectionList> connList(new ConnectionList());

	// Run full system scan
	std::cout << "Scanning for connected iMS Systems. Please wait. . . ." << std::endl;
	std::vector<IMSSystem> fulliMSList = connList->scan();

	auto myiMS = std::make_shared<IMSSystem>();
	if (fulliMSList.empty()) {
		std::cout << "Can't connect to an iMS System.  Press a key to exit." << std::endl;
		while (!_kbhit());
		return -1;
	}
	else {
		std::cout << "Discovered " << fulliMSList.size() << " iMS Systems: " << std::endl;
		int ims_count = 1;
		for (std::vector<IMSSystem>::const_iterator it = fulliMSList.cbegin(); it != fulliMSList.cend(); ++it) {
			std::cout << "  " << ims_count++ << ") " << it->ConnPort() << std::endl;
		}
		while (1) {
			std::cout << "Select a system to continue ... ";
			char ch;
			std::cin >> ch;
			// Don't connect more than 9 systems!
			if ((ch - '0') > 0 && (ch - '0') <= (int)(fulliMSList.size())) {
				*myiMS = fulliMSList[ch - '1'];
				break;
			}
		}
		std::cout << std::endl;
	}

	std::cout << "Connecting ... ";
	myiMS->Connect();
	std::cout << "OK" << std::endl;

	std::cout << std::endl << "Connecting to IMS System on port: " << myiMS->ConnPort() << std::endl;
	if (myiMS->Synth().IsValid()) {
		std::cout << "Found Synthesiser: " << myiMS->Synth().Model() << " : " << myiMS->Synth().Description() << std::endl;
		std::cout << " FW Version: " << myiMS->Synth().GetVersion() << std::endl;
	}
	if (myiMS->Ctlr().IsValid()) {
		std::cout << "Found Controller: " << myiMS->Ctlr().Model() << " : " << myiMS->Ctlr().Description() << std::endl;
		std::cout << " FW Version: " << myiMS->Ctlr().GetVersion() << std::endl;
	}
	std::cout << std::endl;

	// Create global access object
	glbl = std::make_unique<MyIMSAccess>(myiMS);

	setInitialConditions();

	ImageGroup LissajousGroup;

	const int min_img = 5;
	const int max_img = 32768;
	int n_img = TEST_NUMBER_OF_IMAGES;
	n_img = (n_img > max_img) ? max_img : (n_img < min_img) ? min_img : n_img;

	const int tot_pts = 2500000;

	int pts_per_img = tot_pts / n_img;

	// Builds an image table of up to 32K in size
	//
	// Image Table index:
	// bits [4:0] = lissajous phase
	// bits [7:5] = lissajous X order
	// bits [10:8] = lissajous Y order
	// bits [12:11] = Horizontal location
	// bits [14:13] = Vertical location
	//
	int img_count = 0;
	do
	{
		for (int y = 0; y < 4 && img_count < n_img; y++)
		{
			for (int x = 3; x >= 0 && img_count < n_img; x--)
			{
				for (int xinc = 1; xinc <= 8 && img_count < n_img; xinc++)
				{
					for (int yinc = 1; yinc <= 8 && img_count < n_img; yinc++)
					{
						for (int phase = 0; phase <= 31 && img_count < n_img; phase++, img_count++)
						{
							LissajousImageGen liss;
							liss.SetCentre(50 + x * 40, 50 + y * 40);
							liss.SetSpan(18);

							liss.ConfigureLissajous(Degrees(phase * (180.0 / 32)), xinc, yinc);
							auto img = liss.Generate(pts_per_img);

							img->ClockRate(kHz(ImgClk));
							LissajousGroup.AddImage(*img);
						}
					}
				}
			}
		}

	} while (0);

	downloadImagesToIMS(myiMS, LissajousGroup);

	glbl->seqMgr().Stop(ImagePlayer::StopStyle::IMMEDIATELY);

	// Build Sequences
	const int min_seq = 1;
	const int max_seq = 10;
	int n_seq = TEST_NUMBER_OF_SEQUENCES;
	n_seq = (n_seq > max_seq) ? max_seq : (n_seq < min_seq) ? min_seq : n_seq;

	const int max_seq_entries = 3000000;

	int seq_entries_per_seq = max_seq_entries / n_seq;

	std::deque<ImageSequence> seqArray;

	// seed random number generator
	srand((unsigned)time(NULL));

	for (int seqcnt = 0; seqcnt < n_seq; seqcnt++)
	{
		seqArray.push_back(ImageSequence(SequenceTermAction::DISCARD));

		// Random selection of lissajous figure / location within boundaries of index table
		int liss_index = rand() % LissajousGroup.Size();

		// Start at phase 0
		liss_index &= ~0x1F;

		const int max_phase_divisor = (500000 / pts_per_img);
		for (int k = 0; k < seq_entries_per_seq; k++)
		{
			// cycle through 32 phases spread over seq_entries
			int phase_divisor = (seq_entries_per_seq < max_phase_divisor) ? seq_entries_per_seq : max_phase_divisor;
			auto index = liss_index + (((32 * k) / phase_divisor) % 32);
			if (index >= LissajousGroup.Size()) index = LissajousGroup.Size() - 1;
			seqArray[seqcnt].push_back(std::make_shared<ImageSequenceEntry>(LissajousGroup[index], ImageRepeats::NONE));
		}
	}

	glbl->seqMgr().SequenceEventSubscribe(SequenceEvents::SEQUENCE_ERROR, &seqs);
	glbl->seqMgr().SequenceEventSubscribe(SequenceEvents::SEQUENCE_START, &seqs);
	glbl->seqMgr().SequenceEventSubscribe(SequenceEvents::SEQUENCE_FINISHED, &seqs);
	glbl->seqMgr().SequenceEventSubscribe(SequenceEvents::SEQUENCE_POSITION, &seqs);

	char ch;
	do {
		// Clear Queue and Download Sequences
		glbl->seqMgr().QueueClear();

		for (int seqcnt = 0; seqcnt < n_seq; seqcnt++)
		{
			std::cout << "Downloading Sequence " << seqcnt << " [" << seqArray[seqcnt].size() << " entries]" << std::endl;
			wait_for_sequence_download(*myiMS, seqArray[seqcnt]);
		}

		// Run through sequences
		glbl->seqMgr().StartSequenceQueue(SequenceManager::SeqConfiguration(SequenceManager::PointClock::EXTERNAL, SequenceManager::ImageTrigger::CONTINUOUS),
			SequenceManager::ImageTrigger::CONTINUOUS);

		do {
			std::this_thread::sleep_for(std::chrono::milliseconds(1000));
		} while (seqs.Playing());

		std::cout << "Rerun sequence test? [y/n] " << std::endl;
		std::cin >> ch;
	} while (ch != 'n' && ch != 'N');

	glbl->seqMgr().SequenceEventUnsubscribe(SequenceEvents::SEQUENCE_ERROR, &seqs);
	glbl->seqMgr().SequenceEventUnsubscribe(SequenceEvents::SEQUENCE_START, &seqs);
	glbl->seqMgr().SequenceEventUnsubscribe(SequenceEvents::SEQUENCE_FINISHED, &seqs);
	glbl->seqMgr().SequenceEventUnsubscribe(SequenceEvents::SEQUENCE_POSITION, &seqs);

	myiMS->Disconnect();
	glbl.release();

	return 0;
}

